@@ -1,5 +1,5 @@
 # sql/types_api.py
-# Copyright (C) 2005-2016 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -10,10 +10,13 @@
 """
 
 
-from .. import exc, util
 from . import operators
-from .visitors import Visitable, VisitableType
 from .base import SchemaEventTarget
+from .visitors import Visitable
+from .visitors import VisitableType
+from .. import exc
+from .. import util
+
 
 # these are back-assigned by sqltypes.
 BOOLEANTYPE = None
@@ -49,7 +52,8 @@ class TypeEngine(Visitable):
 
 
         """
-        __slots__ = 'expr', 'type'
+
+        __slots__ = "expr", "type"
 
         default_comparator = None
 
@@ -57,16 +61,15 @@ class TypeEngine(Visitable):
             self.expr = expr
             self.type = expr.type
 
-        @util.dependencies('sqlalchemy.sql.default_comparator')
+        @util.dependencies("sqlalchemy.sql.default_comparator")
         def operate(self, default_comparator, op, *other, **kwargs):
             o = default_comparator.operator_lookup[op.__name__]
             return o[0](self.expr, op, *(other + o[1:]), **kwargs)
 
-        @util.dependencies('sqlalchemy.sql.default_comparator')
+        @util.dependencies("sqlalchemy.sql.default_comparator")
         def reverse_operate(self, default_comparator, op, other, **kwargs):
             o = default_comparator.operator_lookup[op.__name__]
-            return o[0](self.expr, op, other,
-                        reverse=True, *o[1:], **kwargs)
+            return o[0](self.expr, op, other, reverse=True, *o[1:], **kwargs)
 
         def _adapt_expression(self, op, other_comparator):
             """evaluate the return type of <self> <op> <othertype>,
@@ -74,13 +77,14 @@ class TypeEngine(Visitable):
 
             This method determines the type of a resulting binary expression
             given two source types and an operator.   For example, two
-            :class:`.Column` objects, both of the type :class:`.Integer`, will
+            :class:`_schema.Column` objects, both of the type
+            :class:`.Integer`, will
             produce a :class:`.BinaryExpression` that also has the type
             :class:`.Integer` when compared via the addition (``+``) operator.
             However, using the addition operator with an :class:`.Integer`
             and a :class:`.Date` object will produce a :class:`.Date`, assuming
             "days delta" behavior by the database (in reality, most databases
-            other than Postgresql don't accept this particular operation).
+            other than PostgreSQL don't accept this particular operation).
 
             The method returns a tuple of the form <operator>, <type>.
             The resulting operator and type will be those applied to the
@@ -93,10 +97,11 @@ class TypeEngine(Visitable):
             boolean comparison or special SQL keywords like MATCH or BETWEEN.
 
             """
+
             return op, self.type
 
         def __reduce__(self):
-            return _reconstitute_comparator, (self.expr, )
+            return _reconstitute_comparator, (self.expr,)
 
     hashable = True
     """Flag, if False, means values from this type aren't hashable.
@@ -107,7 +112,8 @@ class TypeEngine(Visitable):
 
     comparator_factory = Comparator
     """A :class:`.TypeEngine.Comparator` class which will apply
-    to operations performed by owning :class:`.ColumnElement` objects.
+    to operations performed by owning :class:`_expression.ColumnElement`
+    objects.
 
     The :attr:`.comparator_factory` attribute is a hook consulted by
     the core expression system when column and SQL expression operations
@@ -126,8 +132,15 @@ class TypeEngine(Visitable):
     of existing types, or alternatively by using :class:`.TypeDecorator`.
     See the documentation section :ref:`types_operators` for examples.
 
-    .. versionadded:: 0.8  The expression system was enhanced to support
-      customization of operators on a per-type level.
+    """
+
+    sort_key_function = None
+    """A sorting function that can be passed as the key to sorted.
+
+    The default value of ``None`` indicates that the values stored by
+    this type are self-sorting.
+
+    .. versionadded:: 1.3.8
 
     """
 
@@ -179,16 +192,24 @@ class TypeEngine(Visitable):
 
         In all cases, the actual NULL SQL value can be always be
         persisted in any column by using
-        the :obj:`~.expression.null` SQL construct in an INSERT statement
+        the :obj:`_expression.null` SQL construct in an INSERT statement
         or associated with an ORM-mapped attribute.
 
+        .. note::
+
+            The "evaluates none" flag does **not** apply to a value
+            of ``None`` passed to :paramref:`_schema.Column.default` or
+            :paramref:`_schema.Column.server_default`; in these cases,
+            ``None``
+            still means "no default".
+
         .. versionadded:: 1.1
 
         .. seealso::
 
             :ref:`session_forcing_null` - in the ORM documentation
 
-            :paramref:`.postgresql.JSON.none_as_null` - Postgresql JSON
+            :paramref:`.postgresql.JSON.none_as_null` - PostgreSQL JSON
             interaction with this flag.
 
             :attr:`.TypeEngine.should_evaluate_none` - class-level flag
@@ -210,7 +231,7 @@ class TypeEngine(Visitable):
         where it can be consumed by schema comparison tools such as
         Alembic autogenerate.
 
-        A future release of SQLAlchemy will potentially impement this method
+        A future release of SQLAlchemy will potentially implement this method
         for builtin types as well.
 
         The function should return True if this type is equivalent to the
@@ -288,9 +309,9 @@ class TypeEngine(Visitable):
         The method is evaluated at statement compile time, as opposed
         to statement construction time.
 
-        See also:
+        .. seealso::
 
-        :ref:`types_sql_value_processing`
+            :ref:`types_sql_value_processing`
 
         """
 
@@ -305,11 +326,13 @@ class TypeEngine(Visitable):
 
         """
 
-        return self.__class__.column_expression.__code__ \
+        return (
+            self.__class__.column_expression.__code__
             is not TypeEngine.column_expression.__code__
+        )
 
     def bind_expression(self, bindvalue):
-        """"Given a bind value (i.e. a :class:`.BindParameter` instance),
+        """Given a bind value (i.e. a :class:`.BindParameter` instance),
         return a SQL expression in its place.
 
         This is typically a SQL function that wraps the existing bound
@@ -327,9 +350,9 @@ class TypeEngine(Visitable):
         may be used in an executemany() call against an arbitrary number
         of bound parameter sets.
 
-        See also:
+        .. seealso::
 
-        :ref:`types_sql_value_processing`
+            :ref:`types_sql_value_processing`
 
         """
         return None
@@ -343,8 +366,14 @@ class TypeEngine(Visitable):
 
         """
 
-        return self.__class__.bind_expression.__code__ \
+        return (
+            self.__class__.bind_expression.__code__
             is not TypeEngine.bind_expression.__code__
+        )
+
+    @staticmethod
+    def _to_instance(cls_or_self):
+        return to_instance(cls_or_self)
 
     def compare_values(self, x, y):
         """Compare two values for equality."""
@@ -355,7 +384,7 @@ class TypeEngine(Visitable):
         """Return the corresponding type object from the underlying DB-API, if
         any.
 
-         This can be useful for calling ``setinputsizes()``, for example.
+        This can be useful for calling ``setinputsizes()``, for example.
 
         """
         return None
@@ -380,7 +409,7 @@ class TypeEngine(Visitable):
         raise NotImplementedError()
 
     def with_variant(self, type_, dialect_name):
-        """Produce a new type object that will utilize the given
+        r"""Produce a new type object that will utilize the given
         type when applied to the dialect of the given name.
 
         e.g.::
@@ -398,14 +427,12 @@ class TypeEngine(Visitable):
         itself provides a :meth:`.Variant.with_variant`
         that can be called repeatedly.
 
-        :param type_: a :class:`.TypeEngine` that will be selected
+        :param type\_: a :class:`.TypeEngine` that will be selected
          as a variant from the originating type, when a dialect
          of the given name is in use.
         :param dialect_name: base name of the dialect which uses
          this type. (i.e. ``'postgresql'``, ``'mysql'``, etc.)
 
-        .. versionadded:: 0.7.2
-
         """
         return Variant(self, {dialect_name: to_instance(type_)})
 
@@ -429,28 +456,48 @@ class TypeEngine(Visitable):
 
         """
         try:
-            return dialect._type_memos[self]['impl']
+            return dialect._type_memos[self]["impl"]
         except KeyError:
-            return self._dialect_info(dialect)['impl']
+            return self._dialect_info(dialect)["impl"]
+
+    def _unwrapped_dialect_impl(self, dialect):
+        """Return the 'unwrapped' dialect impl for this type.
+
+        For a type that applies wrapping logic (e.g. TypeDecorator), give
+        us the real, actual dialect-level type that is used.
+
+        This is used by TypeDecorator itself as well at least one case where
+        dialects need to check that a particular specific dialect-level
+        type is in use, within the :meth:`.DefaultDialect.set_input_sizes`
+        method.
+
+        """
+        return self.dialect_impl(dialect)
 
     def _cached_literal_processor(self, dialect):
         """Return a dialect-specific literal processor for this type."""
         try:
-            return dialect._type_memos[self]['literal']
+            return dialect._type_memos[self]["literal"]
         except KeyError:
-            d = self._dialect_info(dialect)
-            d['literal'] = lp = d['impl'].literal_processor(dialect)
-            return lp
+            pass
+        # avoid KeyError context coming into literal_processor() function
+        # raises
+        d = self._dialect_info(dialect)
+        d["literal"] = lp = d["impl"].literal_processor(dialect)
+        return lp
 
     def _cached_bind_processor(self, dialect):
         """Return a dialect-specific bind processor for this type."""
 
         try:
-            return dialect._type_memos[self]['bind']
+            return dialect._type_memos[self]["bind"]
         except KeyError:
-            d = self._dialect_info(dialect)
-            d['bind'] = bp = d['impl'].bind_processor(dialect)
-            return bp
+            pass
+        # avoid KeyError context coming into bind_processor() function
+        # raises
+        d = self._dialect_info(dialect)
+        d["bind"] = bp = d["impl"].bind_processor(dialect)
+        return bp
 
     def _cached_result_processor(self, dialect, coltype):
         """Return a dialect-specific result processor for this type."""
@@ -458,12 +505,27 @@ class TypeEngine(Visitable):
         try:
             return dialect._type_memos[self][coltype]
         except KeyError:
-            d = self._dialect_info(dialect)
-            # key assumption: DBAPI type codes are
-            # constants.  Else this dictionary would
-            # grow unbounded.
-            d[coltype] = rp = d['impl'].result_processor(dialect, coltype)
-            return rp
+            pass
+        # avoid KeyError context coming into result_processor() function
+        # raises
+        d = self._dialect_info(dialect)
+        # key assumption: DBAPI type codes are
+        # constants.  Else this dictionary would
+        # grow unbounded.
+        d[coltype] = rp = d["impl"].result_processor(dialect, coltype)
+        return rp
+
+    def _cached_custom_processor(self, dialect, key, fn):
+        try:
+            return dialect._type_memos[self][key]
+        except KeyError:
+            pass
+        # avoid KeyError context coming into fn() function
+        # raises
+        d = self._dialect_info(dialect)
+        impl = d["impl"]
+        d[key] = result = fn(impl)
+        return result
 
     def _dialect_info(self, dialect):
         """Return a dialect-specific registry which
@@ -478,7 +540,7 @@ class TypeEngine(Visitable):
                 impl = self.adapt(type(self))
             # this can't be self, else we create a cycle
             assert impl is not self
-            dialect._type_memos[self] = d = {'impl': impl}
+            dialect._type_memos[self] = d = {"impl": impl}
             return d
 
     def _gen_dialect_impl(self, dialect):
@@ -514,8 +576,10 @@ class TypeEngine(Visitable):
 
         """
         _coerced_type = _resolve_value_to_type(value)
-        if _coerced_type is NULLTYPE or _coerced_type._type_affinity \
-                is self._type_affinity:
+        if (
+            _coerced_type is NULLTYPE
+            or _coerced_type._type_affinity is self._type_affinity
+        ):
             return self
         else:
             return _coerced_type
@@ -551,8 +615,9 @@ class TypeEngine(Visitable):
 
     def __str__(self):
         if util.py2k:
-            return unicode(self.compile()).\
-                encode('ascii', 'backslashreplace')
+            return unicode(self.compile()).encode(  # noqa
+                "ascii", "backslashreplace"
+            )  # noqa
         else:
             return str(self.compile())
 
@@ -599,7 +664,7 @@ class UserDefinedType(util.with_metaclass(VisitableCheckKWArg, TypeEngine)):
 
     The ``get_col_spec()`` method will in most cases receive a keyword
     argument ``type_expression`` which refers to the owning expression
-    of the type as being compiled, such as a :class:`.Column` or
+    of the type as being compiled, such as a :class:`_schema.Column` or
     :func:`.cast` construct.  This keyword is only sent if the method
     accepts keyword arguments (e.g. ``**kw``) in its argument signature;
     introspection is used to check for this in order to support legacy
@@ -610,15 +675,16 @@ class UserDefinedType(util.with_metaclass(VisitableCheckKWArg, TypeEngine)):
        ``type_expression``, if it receives ``**kw`` in its signature.
 
     """
+
     __visit_name__ = "user_defined"
 
-    ensure_kwarg = 'get_col_spec'
+    ensure_kwarg = "get_col_spec"
 
     class Comparator(TypeEngine.Comparator):
         __slots__ = ()
 
         def _adapt_expression(self, op, other_comparator):
-            if hasattr(self.type, 'adapt_operator'):
+            if hasattr(self.type, "adapt_operator"):
                 util.warn_deprecated(
                     "UserDefinedType.adapt_operator is deprecated.  Create "
                     "a UserDefinedType.Comparator subclass instead which "
@@ -627,7 +693,9 @@ class UserDefinedType(util.with_metaclass(VisitableCheckKWArg, TypeEngine)):
                 )
                 return self.type.adapt_operator(op), self.type
             else:
-                return op, self.type
+                return super(
+                    UserDefinedType.Comparator, self
+                )._adapt_expression(op, other_comparator)
 
     comparator_factory = Comparator
 
@@ -640,16 +708,78 @@ class UserDefinedType(util.with_metaclass(VisitableCheckKWArg, TypeEngine)):
         the same type as this one.  See
         :meth:`.TypeDecorator.coerce_compared_value` for more detail.
 
-        .. versionchanged:: 0.8 :meth:`.UserDefinedType.coerce_compared_value`
-           now returns ``self`` by default, rather than falling onto the
-           more fundamental behavior of
-           :meth:`.TypeEngine.coerce_compared_value`.
-
         """
 
         return self
 
 
+class Emulated(object):
+    """Mixin for base types that emulate the behavior of a DB-native type.
+
+    An :class:`.Emulated` type will use an available database type
+    in conjunction with Python-side routines and/or database constraints
+    in order to approximate the behavior of a database type that is provided
+    natively by some backends.  When a native-providing backend is in
+    use, the native version of the type is used.  This native version
+    should include the :class:`.NativeForEmulated` mixin to allow it to be
+    distinguished from :class:`.Emulated`.
+
+    Current examples of :class:`.Emulated` are:  :class:`.Interval`,
+    :class:`.Enum`, :class:`.Boolean`.
+
+    .. versionadded:: 1.2.0b3
+
+    """
+
+    def adapt_to_emulated(self, impltype, **kw):
+        """Given an impl class, adapt this type to the impl assuming "emulated".
+
+        The impl should also be an "emulated" version of this type,
+        most likely the same class as this type itself.
+
+        e.g.: sqltypes.Enum adapts to the Enum class.
+
+        """
+        return super(Emulated, self).adapt(impltype, **kw)
+
+    def adapt(self, impltype, **kw):
+        if hasattr(impltype, "adapt_emulated_to_native"):
+
+            if self.native:
+                # native support requested, dialect gave us a native
+                # implementor, pass control over to it
+                return impltype.adapt_emulated_to_native(self, **kw)
+            else:
+                # impltype adapts to native, and we are not native,
+                # so reject the impltype in favor of "us"
+                impltype = self.__class__
+
+        if issubclass(impltype, self.__class__):
+            return self.adapt_to_emulated(impltype, **kw)
+        else:
+            return super(Emulated, self).adapt(impltype, **kw)
+
+
+class NativeForEmulated(object):
+    """Indicates DB-native types supported by an :class:`.Emulated` type.
+
+    .. versionadded:: 1.2.0b3
+
+    """
+
+    @classmethod
+    def adapt_emulated_to_native(cls, impl, **kw):
+        """Given an impl, adapt this type's class to the impl assuming "native".
+
+        The impl will be an :class:`.Emulated` class but not a
+        :class:`.NativeForEmulated`.
+
+        e.g.: postgresql.ENUM produces a type given an Enum instance.
+
+        """
+        return cls(**kw)
+
+
 class TypeDecorator(SchemaEventTarget, TypeEngine):
     """Allows the creation of types which add additional functionality
     to an existing type.
@@ -678,10 +808,10 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
           def copy(self, **kw):
               return MyType(self.impl.length)
 
-    The class-level "impl" attribute is required, and can reference any
-    TypeEngine class.  Alternatively, the load_dialect_impl() method
-    can be used to provide different type classes based on the dialect
-    given; in this case, the "impl" variable can reference
+    The class-level ``impl`` attribute is required, and can reference any
+    :class:`.TypeEngine` class.  Alternatively, the :meth:`load_dialect_impl`
+    method can be used to provide different type classes based on the dialect
+    given; in this case, the ``impl`` variable can reference
     ``TypeEngine`` as a placeholder.
 
     Types that receive a Python type that isn't similar to the ultimate type
@@ -736,7 +866,7 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
        If the :class:`.TypeDecorator` is augmenting a
        type that requires special logic for certain types of operators,
        this method **must** be overridden.  A key example is when decorating
-       the :class:`.postgresql.JSON` and :class:`.postgresql.JSONB` types;
+       the :class:`_postgresql.JSON` and :class:`_postgresql.JSONB` types;
        the default rules of :meth:`.TypeEngine.coerce_compared_value` should
        be used in order to deal with operators like index operations::
 
@@ -771,17 +901,19 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
 
         """
 
-        if not hasattr(self.__class__, 'impl'):
-            raise AssertionError("TypeDecorator implementations "
-                                 "require a class-level variable "
-                                 "'impl' which refers to the class of "
-                                 "type being decorated")
+        if not hasattr(self.__class__, "impl"):
+            raise AssertionError(
+                "TypeDecorator implementations "
+                "require a class-level variable "
+                "'impl' which refers to the class of "
+                "type being decorated"
+            )
         self.impl = to_instance(self.__class__.impl, *args, **kwargs)
 
-    coerce_to_is_types = (util.NoneType, )
+    coerce_to_is_types = (util.NoneType,)
     """Specify those Python types which should be coerced at the expression
     level to "IS <constant>" when compared using ``==`` (and same for
-    ``IS NOT`` in conjunction with ``!=``.
+    ``IS NOT`` in conjunction with ``!=``).
 
     For most SQLAlchemy types, this includes ``NoneType``, as well as
     ``bool``.
@@ -793,34 +925,42 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
     return an empty tuple, in which case no values will be coerced to
     constants.
 
-    .. versionadded:: 0.8.2
-        Added :attr:`.TypeDecorator.coerce_to_is_types` to allow for easier
-        control of ``__eq__()`` ``__ne__()`` operations.
-
     """
 
     class Comparator(TypeEngine.Comparator):
+        """A :class:`.TypeEngine.Comparator` that is specific to
+        :class:`.TypeDecorator`.
+
+        User-defined :class:`.TypeDecorator` classes should not typically
+        need to modify this.
+
+
+        """
+
         __slots__ = ()
 
         def operate(self, op, *other, **kwargs):
-            kwargs['_python_is_types'] = self.expr.type.coerce_to_is_types
+            kwargs["_python_is_types"] = self.expr.type.coerce_to_is_types
             return super(TypeDecorator.Comparator, self).operate(
-                op, *other, **kwargs)
+                op, *other, **kwargs
+            )
 
         def reverse_operate(self, op, other, **kwargs):
-            kwargs['_python_is_types'] = self.expr.type.coerce_to_is_types
+            kwargs["_python_is_types"] = self.expr.type.coerce_to_is_types
             return super(TypeDecorator.Comparator, self).reverse_operate(
-                op, other, **kwargs)
+                op, other, **kwargs
+            )
 
     @property
     def comparator_factory(self):
         if TypeDecorator.Comparator in self.impl.comparator_factory.__mro__:
             return self.impl.comparator_factory
         else:
-            return type("TDComparator",
-                        (TypeDecorator.Comparator,
-                         self.impl.comparator_factory),
-                        {})
+            return type(
+                "TDComparator",
+                (TypeDecorator.Comparator, self.impl.comparator_factory),
+                {},
+            )
 
     def _gen_dialect_impl(self, dialect):
         """
@@ -833,13 +973,14 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         # otherwise adapt the impl type, link
         # to a copy of this TypeDecorator and return
         # that.
-        typedesc = self.load_dialect_impl(dialect).dialect_impl(dialect)
+        typedesc = self._unwrapped_dialect_impl(dialect)
         tt = self.copy()
         if not isinstance(tt, self.__class__):
-            raise AssertionError('Type object %s does not properly '
-                                 'implement the copy() method, it must '
-                                 'return an object of type %s' %
-                                 (self, self.__class__))
+            raise AssertionError(
+                "Type object %s does not properly "
+                "implement the copy() method, it must "
+                "return an object of type %s" % (self, self.__class__)
+            )
         tt.impl = typedesc
         return tt
 
@@ -851,13 +992,17 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         return self.impl._type_affinity
 
     def _set_parent(self, column):
-        """Support SchemaEentTarget"""
+        """Support SchemaEventTarget"""
+
+        super(TypeDecorator, self)._set_parent(column)
 
         if isinstance(self.impl, SchemaEventTarget):
             self.impl._set_parent(column)
 
     def _set_parent_with_dispatch(self, parent):
-        """Support SchemaEentTarget"""
+        """Support SchemaEventTarget"""
+
+        super(TypeDecorator, self)._set_parent_with_dispatch(parent)
 
         if isinstance(self.impl, SchemaEventTarget):
             self.impl._set_parent_with_dispatch(parent)
@@ -896,6 +1041,20 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         """
         return self.impl
 
+    def _unwrapped_dialect_impl(self, dialect):
+        """Return the 'unwrapped' dialect impl for this type.
+
+        For a type that applies wrapping logic (e.g. TypeDecorator), give
+        us the real, actual dialect-level type that is used.
+
+        This is used by TypeDecorator itself as well at least one case where
+        dialects need to check that a particular specific dialect-level
+        type is in use, within the :meth:`.DefaultDialect.set_input_sizes`
+        method.
+
+        """
+        return self.load_dialect_impl(dialect).dialect_impl(dialect)
+
     def __getattr__(self, key):
         """Proxy all other undefined accessors to the underlying
         implementation."""
@@ -978,8 +1137,10 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
 
         """
 
-        return self.__class__.process_bind_param.__code__ \
+        return (
+            self.__class__.process_bind_param.__code__
             is not TypeDecorator.process_bind_param.__code__
+        )
 
     @util.memoized_property
     def _has_literal_processor(self):
@@ -988,8 +1149,10 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
 
         """
 
-        return self.__class__.process_literal_param.__code__ \
+        return (
+            self.__class__.process_literal_param.__code__
             is not TypeDecorator.process_literal_param.__code__
+        )
 
     def literal_processor(self, dialect):
         """Provide a literal processing function for the given
@@ -1026,9 +1189,12 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         if process_param:
             impl_processor = self.impl.literal_processor(dialect)
             if impl_processor:
+
                 def process(value):
                     return impl_processor(process_param(value, dialect))
+
             else:
+
                 def process(value):
                     return process_param(value, dialect)
 
@@ -1059,10 +1225,12 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
             process_param = self.process_bind_param
             impl_processor = self.impl.bind_processor(dialect)
             if impl_processor:
+
                 def process(value):
                     return impl_processor(process_param(value, dialect))
 
             else:
+
                 def process(value):
                     return process_param(value, dialect)
 
@@ -1079,8 +1247,10 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         exception throw.
 
         """
-        return self.__class__.process_result_value.__code__ \
+        return (
+            self.__class__.process_result_value.__code__
             is not TypeDecorator.process_result_value.__code__
+        )
 
     def result_processor(self, dialect, coltype):
         """Provide a result value processing function for the given
@@ -1104,13 +1274,14 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         """
         if self._has_result_processor:
             process_value = self.process_result_value
-            impl_processor = self.impl.result_processor(dialect,
-                                                        coltype)
+            impl_processor = self.impl.result_processor(dialect, coltype)
             if impl_processor:
+
                 def process(value):
                     return process_value(impl_processor(value), dialect)
 
             else:
+
                 def process(value):
                     return process_value(value, dialect)
 
@@ -1118,6 +1289,33 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         else:
             return self.impl.result_processor(dialect, coltype)
 
+    @util.memoized_property
+    def _has_bind_expression(self):
+        return (
+            self.__class__.bind_expression.__code__
+            is not TypeDecorator.bind_expression.__code__
+        ) or self.impl._has_bind_expression
+
+    def bind_expression(self, bindparam):
+        return self.impl.bind_expression(bindparam)
+
+    @util.memoized_property
+    def _has_column_expression(self):
+        """memoized boolean, check if column_expression is implemented.
+
+        Allows the method to be skipped for the vast majority of expression
+        types that don't use this feature.
+
+        """
+
+        return (
+            self.__class__.column_expression.__code__
+            is not TypeDecorator.column_expression.__code__
+        ) or self.impl._has_column_expression
+
+    def column_expression(self, column):
+        return self.impl.column_expression(column)
+
     def coerce_compared_value(self, op, value):
         """Suggest a type for a 'coerced' Python value in an expression.
 
@@ -1174,6 +1372,10 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         """
         return self.impl.compare_values(x, y)
 
+    @property
+    def sort_key_function(self):
+        return self.impl.sort_key_function
+
     def __repr__(self):
         return util.generic_repr(self, to_inspect=self.impl)
 
@@ -1185,8 +1387,6 @@ class Variant(TypeDecorator):
     The :class:`.Variant` type is typically constructed
     using the :meth:`.TypeEngine.with_variant` method.
 
-    .. versionadded:: 0.7.2
-
     .. seealso:: :meth:`.TypeEngine.with_variant` for an example of use.
 
     """
@@ -1202,18 +1402,43 @@ class Variant(TypeDecorator):
         self.impl = base
         self.mapping = mapping
 
+    def coerce_compared_value(self, operator, value):
+        result = self.impl.coerce_compared_value(operator, value)
+        if result is self.impl:
+            return self
+        else:
+            return result
+
     def load_dialect_impl(self, dialect):
         if dialect.name in self.mapping:
             return self.mapping[dialect.name]
         else:
             return self.impl
 
+    def _set_parent(self, column):
+        """Support SchemaEventTarget"""
+
+        if isinstance(self.impl, SchemaEventTarget):
+            self.impl._set_parent(column)
+        for impl in self.mapping.values():
+            if isinstance(impl, SchemaEventTarget):
+                impl._set_parent(column)
+
+    def _set_parent_with_dispatch(self, parent):
+        """Support SchemaEventTarget"""
+
+        if isinstance(self.impl, SchemaEventTarget):
+            self.impl._set_parent_with_dispatch(parent)
+        for impl in self.mapping.values():
+            if isinstance(impl, SchemaEventTarget):
+                impl._set_parent_with_dispatch(parent)
+
     def with_variant(self, type_, dialect_name):
-        """Return a new :class:`.Variant` which adds the given
+        r"""Return a new :class:`.Variant` which adds the given
         type + dialect name to the mapping, in addition to the
         mapping present in this :class:`.Variant`.
 
-        :param type_: a :class:`.TypeEngine` that will be selected
+        :param type\_: a :class:`.TypeEngine` that will be selected
          as a variant from the originating type, when a dialect
          of the given name is in use.
         :param dialect_name: base name of the dialect which uses
@@ -1224,7 +1449,8 @@ class Variant(TypeDecorator):
         if dialect_name in self.mapping:
             raise exc.ArgumentError(
                 "Dialect '%s' is already present in "
-                "the mapping for this Variant" % dialect_name)
+                "the mapping for this Variant" % dialect_name
+            )
         mapping = self.mapping.copy()
         mapping[dialect_name] = type_
         return Variant(self.impl, mapping)
@@ -1266,6 +1492,6 @@ def adapt_type(typeobj, colspecs):
     # but it turns out the originally given "generic" type
     # is actually a subclass of our resulting type, then we were already
     # given a more specific type than that required; so use that.
-    if (issubclass(typeobj.__class__, impltype)):
+    if issubclass(typeobj.__class__, impltype):
         return typeobj
     return typeobj.adapt(impltype)
